/*=============================================================================*
 *  Copyright 2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *=============================================================================*/
package org.apache.ws.resource.faults;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.ws.Soap1_1Constants;
import org.apache.ws.util.i18n.Messages;
import org.apache.ws.resource.i18n.MessagesImpl;
import org.apache.ws.resource.i18n.Keys;
import org.apache.ws.resource.faults.v1_2.Faults1_2Constants;

import javax.xml.namespace.QName;
import javax.xml.rpc.soap.SOAPFaultException;
import java.util.Calendar;

/**
 * LOG-DONE
 * The WS-BaseFaults base fault type implemented as a Java exception.
 *
 * @author Ian P. Springer
 */
public class BaseFaultException
   extends SOAPFaultException
{
   public static final Messages MSG = MessagesImpl.getInstance();
   /**
    * DOCUMENT_ME
    */
   private static final QName DEFAULT_NAME =
      new QName( Faults1_2Constants.NSURI_BASEFAULTS_SCHEMA, "BaseFault",
                 Faults1_2Constants.NSPREFIX_BASEFAULTS_SCHEMA );

   /**
    * DOCUMENT_ME
    */
   private static final QName    DEFAULT_FAULT_CODE = Soap1_1Constants.FAULT_CLIENT;
   private Calendar              m_timestamp;
   private EndpointReferenceType m_originator;
   private ErrorCode             m_errorCode;
   private QName                 m_name;
   private String[]              m_description;
   private BaseFaultException[]  m_faultCause;

   /**
    * Constructs a new BaseFaultException with the default name (wsbf:BaseFault). The SOAP fault code will be set to
    * soap:Client and the SOAP fault string will be set to the fault name. Sets the timestamp field to the current time
    * and the SOAP actor and detail fields to null.
    */
   public BaseFaultException(  )
   {
      this( null, null, null );
   }

   /**
    * Constructs a new BaseFaultException with the specified name. The SOAP fault code will be set to soap:Client and
    * the SOAP fault string will be set to the fault name. Sets the timestamp field to the current time and the SOAP
    * actor and detail fields to null.
    *
    * @param name XML element name to use when serializing this fault
    */
   public BaseFaultException( QName name )
   {
      this( name, null, null );
   }

   /**
    * Constructs a new BaseFaultException with the specified name and SOAP fault code. The SOAP fault string will be set
    * to the fault name. Sets the timestamp field to the current time and the SOAP actor and detail fields to null.
    *
    * @param name      XML element name to use when serializing this fault
    * @param faultCode the SOAP fault code
    */
   public BaseFaultException( QName name,
                              QName faultCode )
   {
      this( name, faultCode, null );
   }

   /**
    * Constructs a new BaseFaultException with the specified name, SOAP fault code, and SOAP fault string. Sets the
    * timestamp field to the current time and the SOAP actor and detail fields to null.
    *
    * @param name        XML element name to use when serializing this fault
    * @param faultCode   the SOAP fault code
    * @param faultString the SOAP fault string; if null, String.valueOf( name ) will be used as faultString
    */
   public BaseFaultException( QName  name,
                              QName  faultCode,
                              String faultString )
   {
      super( ( faultCode != null ) ? faultCode : DEFAULT_FAULT_CODE,
             ( faultString != null ) ? faultString : String.valueOf( name ), null, null );
      m_name         = ( name != null ) ? name : DEFAULT_NAME;
      m_timestamp    = Calendar.getInstance(  );
   }

   /**
    * DOCUMENT_ME
    *
    * @param description DOCUMENT_ME
    */
   public void setDescription( String[] description )
   {
      m_description = description;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String[] getDescription(  )
   {
      return m_description;
   }

   /**
    * DOCUMENT_ME
    *
    * @param errorCode DOCUMENT_ME
    */
   public void setErrorCode( ErrorCode errorCode )
   {
      m_errorCode = errorCode;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public ErrorCode getErrorCode(  )
   {
      return m_errorCode;
   }

   /**
    * DOCUMENT_ME
    *
    * @param faultCause DOCUMENT_ME
    */
   public void setFaultCause( BaseFaultException[] faultCause )
   {
      m_faultCause = faultCause;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public BaseFaultException[] getFaultCause(  )
   {
      return m_faultCause;
   }

   /**
    * @return
    */
   public QName getName(  )
   {
      return m_name;
   }

   /**
    * DOCUMENT_ME
    *
    * @param originator DOCUMENT_ME
    */
   public void setOriginator( EndpointReferenceType originator )
   {
      m_originator = originator;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public EndpointReferenceType getOriginator(  )
   {
      return m_originator;
   }

   /**
    * DOCUMENT_ME
    *
    * @param timestamp DOCUMENT_ME
    */
   public void setTimestamp( Calendar timestamp )
   {
      m_timestamp = timestamp;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public Calendar getTimestamp(  )
   {
      return m_timestamp;
   }

   /**
    * DOCUMENT_ME
    *
    * @return DOCUMENT_ME
    */
   public String toString(  )
   {
      return new ToStringBuilder( this ).append( MSG.getMessage(Keys.NAME), m_name ).append( MSG.getMessage(Keys.FAULTCODE),
                                                                          getFaultCode(  ) )
                                        .append( MSG.getMessage(Keys.FAULTSTRING),
                                                 getFaultString(  ) ).append( MSG.getMessage(Keys.DESCRIPTION), m_description )
                                        .append( MSG.getMessage(Keys.ERRORCODE), m_errorCode ).append( MSG.getMessage(Keys.FAULTCAUSE) , m_faultCause )
                                        .toString(  );
   }
}